//************************************************************************
// T6963.C
// TextRoutinen fr LCD-Displays mit T6963
//
// Hier fr Winstar WG24064A
//
// Ausgelegt bisher nur auf diesen Displaytyp
// Alle Variablen zur Koordinatenbergabe sind 8Bit lang.
// Das reicht fr Displays mit 128x128, 240x64 und 240x128.
//
// Der Nullpunkt des Displays liegt oben,links.
//
// hk@holger-klabunde.de
// http://www.holger-klabunde.de
// Compiler MCC18 V2.4
//
// 26.01.2006
// Umgestiegen auf Auto Write in SetPixelArray();
// xpos,ypos werden jetzt global bergeben
// pixbuffer[] wird global bergeben
// pixlen wird global bergeben
// Mehr Macros benutzt. Dient der bersicht.
//
// 24.01.2006 und davor
// Hoch optimierte Version !
// WriteData()
// WriteData1()
// WriteData2()
// WriteCommand()
// CheckBusy()
// in den sehr hufig benutzen Routinen durch Code ersetzt.
//
// SetPixel() komplett entsorgt
//
//************************************************************************

#include <p18cxxx.h>
#include "T6963.h"

unsigned char xcursor,ycursor;
unsigned char cgram[9];
unsigned int gfxbase;
unsigned char xpos,ypos;
unsigned char pixbuffer[96];
unsigned char pixlen;

//unsigned char t6963_busy; //making this global hangs program
                            //because using in a MACRO. use local

/* never used anymore, but read the comments !
   used compiler C18 V2.4
//##########################################################
//StatusBits STA0,1 checken
unsigned char CheckBusy(void)
{
 unsigned char by;

 A0=COMMAND;
 DATA_DIR_IN();
 do{
    RD=0;
    by=RD_DATA();
//    by&=0x03;  //hier haut der Compiler doch glatt sechs BCF Befehle rein !
    RD=1;
//   }while(by!= 3); //s.o.
   }while((by & 0x03) != 3); //so gehts schneller
   
 DATA_DIR_OUT(); //Alles wieder auf Ausgnge

 return (by & 0x03);
}
*/

//##########################################################
//Datenbyte schreiben
void WriteData(unsigned char dat)
{
 unsigned char t6963_busy;

 CHECK_BUSY(CHECK_STA0_STA1);
 WRITE_T6963_DATA(dat);
}

//##########################################################
//Kommandobyte schreiben
void WriteCommand(unsigned char comand)
{
 unsigned char t6963_busy;

 CHECK_BUSY(CHECK_STA0_STA1);
 WRITE_T6963_COMMAND(comand);
}

//###############################################################
//Schreibt ein Datenbyte und ein Kommandobyte
//Bisher ist das Kommandobyte immer 0xC0 fr Write+Autoincrement
//Deshalb keine Parameterbergabe fr 'command' mehr !
//Spart Speicher
void WriteData1(unsigned char dat)
{
 WriteData(dat); //Daten schreiben
 WriteCommand(AUTO_INCREMENT); //Kommando schreiben
}

//##########################################################
//Schreibt zwei Datenbytes und ein Kommandobyte
void WriteData2(unsigned int dat, unsigned char command)
{
 WriteData((unsigned char)(dat)); //Daten schreiben
 WriteData((unsigned char)(dat>>8)); //Daten schreiben
 WriteCommand(command); //Kommando schreiben
}

//##########################################################
//Pixelarray bis zu 8 Bit auf dem Display setzen
void SetPixelArray(void)
//##########################################################
{
 unsigned int adress;
 unsigned char t6963_busy,i;
 unsigned char *p;
 
 p=pixbuffer;
 
//Position im Display-RAM setzen
 adress=(unsigned int)G_BASE; //Start of graphicarea

//anhand von xpos und ypos ermitteln welches Byte
//gelesen/geschrieben werden soll

 adress+= BYTES_PER_ROW * (unsigned int)ypos;
 adress+= xpos / PIXPERBYTE;

//############# Begin set adress ############################################
//Set address pointer
 CHECK_BUSY(CHECK_STA0_STA1);
 WRITE_T6963_DATA((unsigned char)(adress));

 CHECK_BUSY(CHECK_STA0_STA1);
 WRITE_T6963_DATA((unsigned char)(adress>>8));

 CHECK_BUSY(CHECK_STA0_STA1);
 WRITE_T6963_COMMAND(ADRESS_SET);

 CHECK_BUSY(CHECK_STA0_STA1);
 WRITE_T6963_COMMAND(AUTO_WRITE);
   
//############# End set adress ############################################

//Note: Status check STA3 from now on !
//############# Begin write data ############################################
 for(i=0; i<pixlen; i++)
  {
   CHECK_BUSY(CHECK_STA3);
   WRITE_T6963_DATA(*p++);
  }
//############# End write data ############################################

 CHECK_BUSY(CHECK_STA3);
 WRITE_T6963_COMMAND(AUTO_RESET);
}

//##########################################################
void InitDisplay(void)
{
 unsigned char i;

 DATA_DIR_OUT();
 WR_DATA(0);
 
 WR=1; RD=1; A0=DATA;
 WR_TRIS=0;
 RD_TRIS=0;
 A0_TRIS=0;
 
//Write text home address=0x0000
 WriteData2(T_BASE,0x40);

//Text area column set= Setze Textspalten 
 WriteData2(BYTES_PER_ROW,0x41);

//Write graphic home address
 WriteData2(G_BASE,0x42);

//Graphics area column set in Bytes !!!
 WriteData2(BYTES_PER_ROW,0x43);

//Set Offset Register= Setze Startadresse CGRAM
 WriteData(CGRAM_OFFSET);
 WriteData(0);
 WriteCommand(0x22);

//set display mode = OR
 WriteCommand(0x80);

//Text only display :(
// WriteCommand(0x94);

//Text and graphic display !!
 WriteCommand(0x9C);

 ClearScreen();
}

//##########################################################
// Kompletten Bildschirm lschen
void ClearScreen(void)
{
 unsigned int i;
 unsigned char t6963_busy;

//Clear graphic screen
 WriteData2(G_BASE,ADRESS_SET); //Daten schreiben
 for(i=G_BASE; i<=G_END; i++)
  {
   CHECK_BUSY(CHECK_STA0_STA1);
   WRITE_T6963_DATA(0x00);

   CHECK_BUSY(CHECK_STA0_STA1);
   WRITE_T6963_COMMAND(AUTO_INCREMENT);
  } 
 
//Clear text screen
//Write text home address
//Set address pointer = Text home
 LCDPos(1,1);

 for(i=0; i<(COLS * LCD_HEIGHT/8); i++)
  {
   CHECK_BUSY(CHECK_STA0_STA1);
   WRITE_T6963_DATA(0x00);

   CHECK_BUSY(CHECK_STA0_STA1);
   WRITE_T6963_COMMAND(AUTO_INCREMENT);
  } 

//Write text home address
//Set address pointer = Text home
 LCDPos(1,1);
}

//##########################################################
// Zeichen von Windows nach T6963 konvertieren
unsigned char ConvertText(unsigned char ch)
{
 unsigned char tmp;

 tmp=ch;

// if(tmp<=8) { tmp+=0x80; return tmp; } //CGRAM von 0 bis 8 ! 8 Zeichen

 if(tmp<0x20) { tmp+=0x80; return tmp; } //CGRAM von 0 bis 0x1F ! 32 Zeichen
// if(tmp==0xFF) { tmp=0x88; return tmp; } //0xFF heit ausgeflltes Zeichen

//Vom T6963 darstellbare Zeichen
//Das hier vor der bersetzungstabelle knnte noch Probleme geben !
 if(tmp>=0x20 && tmp<=0x7F) { tmp-=0x20; return tmp; }

//bersetzungstabelle fr alle anderen Zeichen
//0xFF ist reserviert als Kennzeichen fr Fehler !!
  switch(tmp)
   {
    case 0xE1: tmp=0x6E; break; // 
//     case 196: tmp=0x6E; break; // 
//     case 228: tmp=0x64; break; // 
    case 0xEF: tmp=0x79; break; // 
//     case 214: tmp=0x79; break; // 
//     case 246: tmp=0x74; break; // 
    case 0xF5: tmp=0x7A; break; // 
//     case 220: tmp=0x7A; break; // 
//     case 252: tmp=0x61; break; // 
    case 0xE2: tmp=0x7E; break; //   naja sieht einigermaen so aus
//     case 223: tmp=223; break; // 
    default: tmp=0xFF; break; //alle anderen Zeichen ungltig
   }

 return tmp;
}

//##########################################################
// Textcursor positionieren an Position zeile,spalte
void LCDPos(unsigned char zeile, unsigned char spalte)
{
 unsigned char by;
 unsigned char t6963_busy;
 unsigned int tmp;

 if(spalte>COLS || zeile>ROWS) return; //Position auerhalb des Displays

 by=zeile-1;
 
 tmp=COLS*(unsigned int)by;
 tmp+=(spalte-1);

 CHECK_BUSY(CHECK_STA0_STA1);
 WRITE_T6963_DATA((unsigned char)(tmp));

 CHECK_BUSY(CHECK_STA0_STA1);
 WRITE_T6963_DATA((unsigned char)(tmp>>8));

 CHECK_BUSY(CHECK_STA0_STA1);
 WRITE_T6963_COMMAND(ADRESS_SET);

 xcursor=spalte;
 ycursor=zeile;
}

//##########################################################
// Einen Buchstaben schreiben.
void LCDChar(unsigned char ch)
{
 unsigned char ps;
 unsigned char t6963_busy;

 ps=ConvertText(ch); //bersetzungstabelle aufrufen
 if(ps!=0xFF)
  {
   //verhindern das auerhalb des Displays geschrieben wird
   if(xcursor<=COLS && ycursor<=ROWS)
    {
     CHECK_BUSY(CHECK_STA0_STA1);
     WRITE_T6963_DATA(ps);

     CHECK_BUSY(CHECK_STA0_STA1);
     WRITE_T6963_COMMAND(AUTO_INCREMENT);

     xcursor++;  //Cursor ein Zeichen weiter
    } 
  
  }//if(ps!=0xFF)

}

//##########################################################
//selbstdefinierte Zeichen ins Character RAM schreiben
//die Zeichen mssen vorher ins Array cgram[] gelegt werden
void MakeCGRam(void)
{
 unsigned char i,by;

 by=cgram[0]*8;

 WriteData(by); //Daten schreiben

 WriteData(CGRAM_START); //Daten schreiben

 WriteCommand(0x24); //Kommando schreiben

 for(i=1; i<9; i++) WriteData1(cgram[i]);

 LCDPos(ycursor,xcursor);  //extrem wichtig ! Textcursor wiederherstellen !
}
